(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

(* This is the main program. *)

(* Set up the parser. This includes tying a recursive knot. *)

let load_and_parse filename =
  let lexbuf =
    LexerUtil.open_in filename
  in
  try
    Parser.program Lexer.main lexbuf
  with Parser.Error ->
    Error.errorb lexbuf "Syntax error."

let () =
  ParserBootstrap.load_and_parse := load_and_parse

(* Examine the command line. *)

let filename =
  ref None

let translate2coq =
  ref false

let usage =
  Printf.sprintf "Usage: %s <filename>" Sys.argv.(0)

let () =
  Arg.parse [
    "--coq", Arg.Set translate2coq, "emit Coq translation on standard output"
  ] (fun name -> filename := Some name) usage

let filename =
  match !filename with
  | Some name ->
      name
  | None ->
      Printf.fprintf stderr "%s\n" usage;
      exit 1

let translate2coq =
  !translate2coq

(* Read the file and parse the program. *)

let program : Syntax.program =
  load_and_parse filename

(* Resolve all identifiers and initialize [Kinds]. *)

let types, lemmas, terms =
  SyntaxImport.program program

(* Make sure that all types are well-kinded. *)

let kinds =
  Types.check types

let types =
  Atom.AtomMap.map (fun (_, ty) -> RawTypes.erase ty) types

let xenv =
  Atom.AtomMap.fold (fun a _ xenv ->
    Export.bind xenv a
  ) types Export.empty

(* Make sure that all lemmas are well-kinded. *)

let () =
  Lemmas.check kinds lemmas

(* Make sure that the lemmas are true. *)

let () =
  Lemmas.play types xenv lemmas

(* Make sure that all terms are well-typed. *)

let tenv =
  Terms.check kinds types xenv terms

(* Print the FORK program as Coq abstract syntax. *)

let () =
  if translate2coq then
    Terms2coq.print terms

