(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

open Printf

(* The kinds in FORK correspond to the types in Nakano's type system. [KStar]
   is the base kind: it classifies types that classify values. [KDefined] is a
   reference to a user-defined kind. [KArrow] is the arrow kind
   constructor. [KLater] is Nakano's modality. *)

type kind =
  | KStar
  | KDefined of Atom.atom
  | KArrow of kind * kind
  | KLater of kind

(* Kinds can be (recursively) defined by the user. A (global) kind
   signature maps atoms to kinds. *)

type signature =
    kind Atom.AtomMap.t

(* ------------------------------------------------------------------------- *)

(* Constructors. *)

let star =
  KStar

let arrow kind1 kind2 =
  KArrow (kind1, kind2)

let rec later k kind =
  if k = 0 then kind else later (k - 1) (KLater kind)

(* In the following comments, I write [k.kind] for [later k kind]. *)

(* ------------------------------------------------------------------------- *)

(* Maps over pairs of kinds. *)

let rec compare kind1 kind2 =
  match kind1, kind2 with
  | KStar, KStar ->
      0
  | KDefined a1, KDefined a2 ->
      Atom.compare a1 a2
  | KArrow (kind11, kind12), KArrow (kind21, kind22) ->
      let c = compare kind11 kind21 in
      if c = 0 then compare kind12 kind22 else c
  | KLater kind1, KLater kind2 ->
      compare kind1 kind2
  | KStar, _ ->
      -1
  | _, KStar ->
      1
  | KDefined _, _ ->
      -1
  | _, KDefined _ ->
      1
  | KArrow _, _ ->
      -1
  | _, KArrow _ ->
      1

let equal kind1 kind2 =
  compare kind1 kind2 = 0

module KindPairMap =
  Map.Make(struct
    type t = kind * kind
    let compare (kind11, kind12) (kind21, kind22) =
      let c = compare kind11 kind21 in
      if c = 0 then compare kind12 kind22 else c
  end)

(* ------------------------------------------------------------------------- *)

(* A global list of all kind definitions. *)

let kinds : signature ref =
  ref Atom.AtomMap.empty

(* Looking up this signature. *)

let definition a =
  Atom.AtomMap.find a !kinds

(* ------------------------------------------------------------------------- *)

(* Displaying atoms. *)

let xenv =
  ref Export.empty

let build_xenv () =
  xenv :=
    Atom.AtomMap.fold (fun a _ env ->
      Export.bind env a
    ) !kinds Export.empty

let export (a : Atom.atom) : string =
  Identifier.name (Export.resolve !xenv a)

(* ------------------------------------------------------------------------- *)

(* Displaying kinds. *)

module P = struct

  open Pprint

  let star =
    text "*"

  let arrow =
    text " ->" ^^ softline

  let later =
    text "later"

  let rec kind0 = function
    | KStar ->
	star
    | KDefined a ->
	text (export a)
    | KLater k ->
	later ^^ space ^^ kind k
    | k ->
	parens (kind k)

  and kind1 = function
    | KArrow (k1, k2) ->
	kind0 k1 ^^ arrow ^^ kind1 k2
    | k ->
	kind0 k

  and kind k =
    kind1 k

  let kind : kind -> string =
    wrap kind

end

let print_kind =
  P.kind

(* ------------------------------------------------------------------------- *)

(* Computing fixed points at type atom to Boolean. *)

module F =
  Fix.Make(Maps.PersistentMapsToImperativeMaps(Atom.AtomMap))(Boolean)

(* The following sanity checks exploit the fact that every infinite path
   through the kinds must pass (infinitely often) at a defined kind. Thus, it
   is sufficient to check that every defined kind is well-formed (resp.
   tail-finite) to ensure that every kind is well-formed (resp. tail-finite). *)

(* ------------------------------------------------------------------------- *)

(* A set of kind definitions is well-formed if and only if every cycle
   crosses a [KLater] constructor. *)

let wf : Atom.atom -> bool =
  F.lfp (fun a ->
    let kind = definition a in
    fun self ->
      let rec wf = function
	| KStar ->
	    true
	| KDefined a ->
	    self a
	| KArrow (kind1, kind2) ->
	    wf kind1 && wf kind2
	| KLater _ ->
	    true
      in
      wf kind
  )

(* Check that this is indeed the case. *)

let check_wf () =
  Atom.AtomMap.iter (fun a _ ->
    if not (wf a) then
      Error.signala [a]
	(sprintf "The kind \"%s\" is ill-formed." (export a))
  ) !kinds

(* ------------------------------------------------------------------------- *)

(* A set of kind definitions is tail-finite if and only if every rightmost
   path ends with [KStar], or, equivalently, if and only if every cycle
   enters the left-hand side of a [KArrow] constructor. *)

let tf : Atom.atom -> bool =
  F.lfp (fun a ->
    let kind = definition a in
    fun self ->
      let rec tf = function
	| KStar ->
	    true
	| KDefined a ->
	    self a
	| KArrow (kind1, kind2) ->
	    tf kind2 (* recursive call on the right-hand side only *)
	| KLater kind ->
	    tf kind
      in
      tf kind
  )

(* Check that this is indeed the case. *)

let check_tf () =
  Atom.AtomMap.iter (fun a _ ->
    if not (tf a) then
      Error.signala [a]
	(sprintf "The kind \"%s\" is not tail-finite." (export a))
  ) !kinds

(* ------------------------------------------------------------------------- *)

(* Installing a global signature. *)

let install ks =
  assert (Atom.AtomMap.is_empty !kinds);
  kinds := ks;
  build_xenv();
  check_wf();
  check_tf();
  Error.signaled()

(* ------------------------------------------------------------------------- *)

(* Checking the subkind relationship. *)

(* The subkind relationship corresponds to Nakano's subtype relationship. *)

(* [subkind kind1 kind2] returns the set of all relative integers [k]
   such that [kind1] is a subkind of [k.kind2]. *)

(* Because [kind] is a subkind of [1.kind], this set is upward closed.
   As a consequence, this set takes one of three forms. It may be the
   set of all relative integers; it may be the set of all integers
   greater than, or equal to, some integer [k]; or it may be empty. *)

(* Note that [k] might be a negative integer. For our specification to
   make sense, the notation [k.kind] must be extended to relative
   integers. When [k] is negative, we consider that [kind1 <= k.kind2]
   holds if and only if [(-k).kind1 <= kind2] holds. *)

(* Although I have not (yet?) formally checked that the use of
   negative integers makes sense, I have proved that [1.kind1 <=
   1.kind2] holds if and only if [kind1 <= kind2] holds. This means
   that moving the origin does not make any difference, and seems
   re-assuring. *)

(* Kinds are recursive: the subkind relationship must be computed as a
   greatest fixed point. We set things up so that the answer to every
   subkind query is initially [Full], and converges down to the
   greatest fixed point. *)

(* This module implements sets of integers of the three forms
   mentioned above, together with a few operations required by the
   greatest fixed point computation. *)

module Set = struct

  type set =
    | Full
    | Above of int
    | Void

  type property =
      set

  let equal set1 set2 =
    match set1, set2 with
    | Void, Void
    | Full, Full ->
	true
    | Above k1, Above k2 ->
	k1 = k2
    | _, _ ->
	false

  (* Because we are interested in computing a greatest fixed point,
     the [bottom] element, which serves as the initial value in the
     fixed point computation, must in fact be the full set. For the
     same reason, the only maximal element is the empty set. *)

  let bottom =
    Full

  let is_maximal = function
    | Void ->
	true
    | _ ->
	false

  (* Translation of a set by some integer offset [i]. *)

  let translate i = function
    | Full ->
	Full
    | Above k ->
	Above (k + i)
    | Void ->
	Void

  let print = function
    | Void ->
	"void"
    | Above k ->
	sprintf "above %d" k
    | Full ->
	"full"

end

open Set

(* We now define the equations that govern the greatest fixed point
   computation.  This takes the form of a function [subkind]. As
   announced above, [subkind self kind1 kind2] must return the set of
   all integers [k] such that [kind1 <= k.kind2] holds. The function
   [subkind] is recursive and works by decomposing the structure of
   [kind1] and [kind2]. When it reaches a user-defined kind, it
   expands the definition and invokes [self]. *)

(* It is not obvious, but it seems straightforward to check, that
   [subkind] is monotone: in short, if [self1] produces larger sets
   than [self2], then [subkind self1] produces larger sets than
   [subkind self2]. (Checking this requires some case analysis,
   especially in the case of arrows.) This guarantees the existence of
   a greatest fixed point. *)

(* The comparison of two arrows cannot return [Full] unless the
   comparison of the codomains has returned [Full]. As a consequence,
   one can prove that, provided the kinds are tail-finite (which they
   are! we have checked this), the greatest fixed point can never be
   [Full]. *)

let rec subkind self kind1 kind2 =
  match kind1, kind2 with

  (* Skip bullets as we go down, and compensate when coming back up. *)

  | KLater kind1, _ ->

      (* Because there is a bullet on the left-hand side, one more bullets
	 must be added to the right-hand side. *)

      translate 1 (subkind self kind1 kind2)

  | _, KLater kind2 ->

      (* Because there already is a bullet on the right-hand side, one fewer
	 bullets need be added to the right-hand side. *)

      translate (-1) (subkind self kind1 kind2)

  (* Upon hitting a user-defined kind, expand it, and continue. *)

  | KDefined a1, _ ->
      self (definition a1, kind2)

  | _, KDefined a2 ->
      self (kind1, definition a2)

  (* At leaves, the problem is trivial. *)

  | KStar, KStar ->

      Above 0

  (* Arrows. *)

  | KArrow (kind11, kind12), KArrow (kind21, kind22) ->

      (* In principle, when comparing two arrows, it is permitted to degrade
	 the left-hand side, that is, to insert extra bullets on the left-hand
	 side. However, this is only useful when a Boolean result is expected
	 -- are these arrows in the subkind relationship? Here, an integer
	 result is expected -- the least number of bullets that must be
	 inserted on the right-hand side. Inserting bullets on the left-hand
	 side would only increase this integer, so it would not help. *)
      
      (* Separately compare the domains and codomains. *)

      begin match subkind self kind21 kind11, subkind self kind12 kind22 with
      | Void, _
      | _, Void ->

	  (* If either of these comparisons fail, fail. *)

	  Void

      | Above k1, Above k2 ->

	  (* The integer [k] that we are looking for must satisfy [k.kind21 <=
	     kind11] and [kind12 <= k.kind22], which are respectively equivalent
	     to [-k >= k1] and [k >= k2]. In other words, [k] must lie in the
	     interval [k2, -k1]. Check that this interval is non-empty. *)

	  (* Because of this check, [subkind k1 k2] is not necessarily the
	     opposite of [subkind k2 k1]. For instance, the interval [0, 1] is
	     non-empty, while the opposite interval [0, -1] is empty. The kind [*
	     -> *] is a subkind of the kind [* -> later *], and [subkind] returns
	     [-1] when applied to these kinds; but the converse relationship
	     fails, as [* -> later *] is not a subkind of [* -> *], regardless of
	     how many bullets are inserted. *)

	  (* If the check succeeds, then we return the least valid value of [k],
	     that is, [k2]. *)

	  if k2 > -k1 then Void else Above k2

      | _, set2 ->

	  (* If either of these sets is full, then the interval emptiness check
	     must succeed. We simply return the right-hand set. *)

	  set2

      end

  (* Structural failure cases. *)

  | KStar, KArrow _
  | KArrow _, KStar ->

      Void

(* Set up the fixed point computation. Perform some currying/uncurrying
   work. *)

module K =
  Fix.Make(Maps.PersistentMapsToImperativeMaps(KindPairMap))(Set)

let debug =
  false

let nesting =
  ref 0

let nest() =
  String.make !nesting ' '

let subkind : kind * kind -> set =
  K.lfp (fun (kind1, kind2) self ->
    if debug then begin
      printf "subkind: %sevaluating: %s <= %s\n%!"
	(nest())
	(print_kind kind1)
	(print_kind kind2);
      nesting := !nesting + 2
    end;
    let result = subkind self kind1 kind2 in
    if debug then begin
      nesting := !nesting - 2;
      printf "subkind: %sgot result: %s\n%!"
	(nest())
	(print result)
    end;
    result
  )

let subkind kind1 kind2 =
  subkind (kind1, kind2)

(* TEMPORARY Testing. *)

let test () =
  let module T = struct

    open Printf

    let bs =
      later 1 star

    let ss =
      arrow star star

    let sbs =
      arrow star bs

    let bsbs =
      arrow bs bs

    let () =
      printf "%s\n%!" (print (subkind ss sbs));
      printf "%s\n%!" (print (subkind bsbs sbs));
      printf "%s\n%!" (print (subkind ss bsbs));
      printf "%s\n%!" (print (subkind (arrow ss star) (arrow bsbs bs)));
      printf "%s\n%!" (print (subkind bsbs ss));
      printf "%s\n%!" (print (subkind (arrow ss star) (arrow bsbs star))); (* fails *)
      ()

    let world =
      ref (Atom.fresh (Identifier.mk "FOO" (0, "foo")))

    let () =
      Atom.AtomMap.iter (fun a _ ->
	if Identifier.name (Atom.identifier a) = "WORLD" then
	  world := a
      ) !kinds

    let world =
      KDefined !world

    let bw =
      later 1 world

    let () =
      printf "%s\n%!" (print (subkind world world));
      printf "%s\n%!" (print (subkind world bw));
      printf "%s\n%!" (print (subkind bw world));
      printf "%s\n%!" (print (subkind world (arrow world (arrow star star))));
      printf "%s\n%!" (print (subkind (arrow world (arrow star star)) world)); (* fails *)
      ()

  end
  in ()

(* TEMPORARY

let install kinds =
  install kinds;
  test ()

*)

(* Define a version of [subkind] that returns a Boolean result. We
   return [true] if and only if [0] is a member of the set that is
   produced by the [subkind] query. *)

let is_subkind kind1 kind2 =
  match subkind kind1 kind2 with
  | Full ->
      assert false
  | Above k ->
      k <= 0
  | Void ->
      false

(* Define a version of [subkind] that returns [true] if [kind1]
   is a subkind of some later version of [kind2]. *)

let is_subkind_later kind1 kind2 =
  match subkind kind1 kind2 with
  | Full ->
      assert false
  | Above _ ->
      true
  | Void ->
      false

(* ------------------------------------------------------------------------- *)

(* Normalize a kind so that an arrow appears at its root. *)

exception KindChecking of Error.location list * string

let rec deconstruct_arrow locs = function
  | KStar ->
      raise (KindChecking (locs, "A type of kind * cannot be applied."))
  | KDefined a ->
      deconstruct_arrow locs (definition a)
  | KArrow (kind1, kind2) ->
      kind1, kind2
  | KLater kind ->
      let kind1, kind2 = deconstruct_arrow locs kind in
      KLater kind1, KLater kind2

(* [kapp locs kind1 kind2] determines what kind is obtained when an object of
   kind [kind1] is applied to an object of kind [kind2]. [KindChecking] is
   raised if such an application is invalid. *)

let kapp locs kind1 kind2 =

  (* [kind1] must be an arrow kind. Obtain its domain and codomain. *)

  let kind11, kind12 = deconstruct_arrow locs kind1 in

  (* Determine how many bullets must be inserted in front of [kind11]
     for [kind2] to be a subkind of [kind11]. *)

  match subkind kind2 kind11 with
  | Full ->
      assert false
  | Above k ->
      if k <= 0 then
	(* No extra bullets are necessary. *)
	kind12
      else
	(* Some bullets are necessary. By inserting them, we degrade
	   both the domain and codomain of the left-hand arrow. *)
	later k kind12
  | Void ->
      (* The actual and formal arguments do not match, regardless
	 of how many bullets are inserted. *)
      raise (KindChecking (locs, sprintf
	"This type application is ill-kinded.\n\
         The actual argument has kind:\n\
         %s\n\
         The formal argument has kind:\n\
         %s\n\
         These kinds do not match (not now, not later)."
	(print_kind kind2)
	(print_kind kind11)
      ))

(* ------------------------------------------------------------------------- *)

(* Determine whether a kind is of the form [later^k *]. *)

let rec is_later_star = function
  | KStar ->
      true
  | KDefined a ->
      is_later_star (definition a)
  | KLater kind ->
      is_later_star kind
  | KArrow _ ->
      false

